import unittest
from greedypermutation import Point, knnsample
from metricspaces import MetricSpace

class TestKNNSample(unittest.TestCase):
    def test_smallinstance(self):
        k = 11
        # S = set(range(0, 100, 4))
        S = set(range(0,100,4)) | set(range(100, 200, 10))
        P = [Point([c]) for c in S]
        M = MetricSpace(P)
        output = list(knnsample(M, k, P[0]))
        print(len(output), output)
        self.assertTrue(len(output) >= len(S) / k)


if __name__ == '__main__':
    unittest.main()



from greedypermutation.point import Point
# from greedypermutation.metricspace import MetricSpace
from greedypermutation.clustergraph import Cluster, ClusterGraph
from greedypermutation.knnsample import knnsample
from greedypermutation.greedytree import GreedyTree



from ds2viz.element import Circle

class VizPoint(Circle):
    def __init__(self, x, y, radius = 1):
        super().__init__(radius)
        self.x, self.y = x, y
        self.align('center', (x,y))

    def __iter__(self):
        yield self.x
        yield self.y

    def dist(self, other):
        return sum((a-b) ** 2 for a,b in zip(self, other)) ** (0.5)

    def __eq__(self, other):
        return list(self) == list(other)

    def __hash__(self):
        return hash((self.x, self.y))



import unittest
from greedypermutation import Point, Cluster, ClusterGraph

class L_inf(tuple):
    def dist(self, other):
        return max(abs(a-b) for a,b in zip(self, other))

class TestCluster(unittest.TestCase):
    def testinit_with_center(self):
        P = [Point([2*i,2*i]) for i in range(100)]
        C = Cluster(Point([99,99]))
        for p in P:
            C.addpoint(p)
        self.assertEqual(len(C), 101)

    def testaddpoint(self):
        a,b,c = Point([1,2]), Point([2,3]), Point([3,4])
        C = Cluster(a)
        self.assertEqual(len(C), 1)
        C.addpoint(b)
        self.assertEqual(len(C), 2)
        C.addpoint(c)
        self.assertEqual(len(C), 3)
        self.assertEqual(C.points, {a,b,c})

    def testaddpoint_duplicatepoint(self):
        a,b = Point([1,2]), Point([2,3])
        C = Cluster(a)
        self.assertEqual(len(C), 1)
        C.addpoint(b)
        self.assertEqual(len(C), 2)
        # Add b a second time.
        C.addpoint(b)
        self.assertEqual(len(C), 2)
        self.assertEqual(C.points, {a,b})
        # Add the center again.
        C.addpoint(a)
        self.assertEqual(len(C), 2)
        self.assertEqual(C.points, {a,b})

    def testupdateradius_empty_cluster(self):
        C = Cluster(Point([1,2,3]))
        C.updateradius()
        self.assertEqual(C.radius, 0)

    def testdist(self):
        A = Cluster(Point([2, 3]))
        self.assertEqual(A.dist(Point([7,3])), 5)
        self.assertEqual(A.dist(A.center), 0)
        self.assertEqual(A.dist(Point([7,15])), 13)

    def testpop(self):
        a,b,c,d = Point([0,0]), Point([100, 0]), Point([0,50]), Point([25,25])
        C = Cluster(a)
        C.addpoint(b)
        C.addpoint(c)
        C.addpoint(d)
        self.assertEqual(C.pop(), b)
        self.assertEqual(C.pop(), c)
        self.assertEqual(C.pop(), d)
        self.assertEqual(C.pop(), None)


    # def testpop_nonuniformdistribution(self):
    #     S = set(range(100)) | set(range(100, 200, 20))
    #     P = [Point([c]) for c in S]
    #     M = MetricSpace(P)
    #     C = [Cluster(P[0])]
    #     for p in P:
    #         C[0].addpoint(p)
    #     for i in range(len(S) - 2)

class TestClusterGraph(unittest.TestCase):
    def testbasicusage(self):
        G = ClusterGraph([Point([i,i]) for i in range(100)])
        # root = next(G.vertices())
        # G.addcluster(Point([100,99]), root)
        # self.assertEqual(len(G), 2)
        # for v in G.vertices():
        #     self.assertEqual(len(set(G.nbrs(v))), 1)

    def testrebalance(self):
        a, b = Point([-1]), Point([200])
        G = ClusterGraph([a,b])
        A = Cluster(a)
        B = Cluster(b)
        for i in range(200):
            B.addpoint(Point([i]))
        self.assertEqual(len(A), 1)
        self.assertEqual(len(B), 201)
        G.rebalance(A, B)
        self.assertEqual(len(B), 101)
        self.assertEqual(len(A), 101)


    def testneighborsofneighborscondition(self):
        """ This somewhat long test was written to expose a bug where
        neighbors of the cluster graph were not properly discovered.
        The construction highlights the need for the cluster graph to be
        undirected.
        """
        a =  L_inf([0, 2, 21, 11, 22, 19])
        aa = L_inf([2, 0, 19, 9, 20, 17])
        b = L_inf([21, 19, 0, 10, 21, 18])
        bb = L_inf([11, 9, 10, 0, 11, 8])
        c = L_inf([22, 20 , 21, 11, 0, 3])
        cc = L_inf([ 19, 17, 18, 8, 3, 0])
        P = [a,aa,b,bb,c,cc]
        G = ClusterGraph(P)
        self.assertEqual(len(G), 1)
        p = G.heap.findmax()
        self.assertEqual(p.center, a)
        self.assertEqual(p.pop(), c)
        G.addcluster(c, p)
        self.assertEqual(len(G), 2)
        p = G.heap.findmax()
        self.assertEqual(p.pop(), b)
        G.addcluster(b, p)
        V = {v.center : v for v in G.vertices()}
        self.assertEqual(set(V), {a,b,c})
        self.assertTrue(V[b] in G.nbrs(V[a]))
        self.assertTrue(V[a] in G.nbrs(V[b]))
        self.assertTrue(V[b] in G.nbrs(V[c]))
        self.assertTrue(V[c] in G.nbrs(V[b]))

        # Before adding aa.
        self.assertTrue(bb in V[b])
        self.assertEqual(V[a].pop(), aa)
        G.addcluster(aa, V[a])
        V = {v.center:v for v in G.vertices()}
        # After adding aa.
        self.assertTrue(cc in V[c])
        self.assertTrue(bb in V[aa])
        self.assertTrue(bb.dist(cc) < bb.dist(aa))
        # That means that we should have an edge from c to aa.
        self.assertTrue(V[aa] in G.nbrs(V[c]))

if __name__ == '__main__':
    unittest.main()



from ds2.priorityqueue import PriorityQueue

class MaxHeap(PriorityQueue):
    def __init__(self, items = (), key = lambda x: x):
        PriorityQueue.__init__(self, items, key = lambda x: - key(x))

    findmax = PriorityQueue.findmin
    removemax = PriorityQueue.removemin



def greedy(M, seed = None, tree = False):
    """
    Return an iterator that yields the points of `M` ordered by a greedy
    permutation.

    The optional `seed` parameter indicates the point that should appear first.

    The `tree` parameter is a flag to set whether or not to return also the
    nearest predecessors in the pairing (thus giving the tree).  If set to
    `True`, the result will iterate over `(point, index)` pairs.
    """
    if tree:
        yield from _greedy(M, seed)
    else:
        for p, i in _greedy(M, seed):
            yield p

def _greedy(M, seed = None):
    """
    Return an iterator that yields `(point, index)` pairs, where `point`
    is the next point in a greedy permutation and `index` is the index of they
    nearest predecessor.

    The optional `seed` parameter indicates the point that should appear first.
    """
    P = list(M)

    # If no seed is provided, use the first point.
    if seed is None:
        seed = P[0]
    else:
        # Put the seed in the first position.
        seed_index = P.index(seed)
        P[0], P[seed_index] = P[seed_index], P[0]
    n = len(P)
    yield P[0], None
    pred = {p:0 for p in P}
    preddist = {p: M.dist(p, P[pred[p]]) for p in P}
    for i in range(1,n):
        farthest = i
        for j in range(i+1, n):
            if preddist[P[j]] > preddist[P[farthest]]:
                farthest  = j
        P[i], P[farthest] = P[farthest], P[i]
        # Update the predecessor distance if necessary.
        for j in range(i+1, n):
            newdistance = M.dist(P[i], P[j])
            if newdistance < preddist[P[j]]:
                pred[P[j]] = i
                preddist[P[j]] = newdistance
        yield P[i], pred[P[i]]



from greedypermutation.clustergraph import Cluster, ClusterGraph
from greedypermutation import GreedyTree

def onehopgreedy(M, seed = None, tree = False):
    """
    Return an iterator that yields the points of `M` ordered by a greedy
    permutation.

    The optional `seed` parameter indicates the point that should appear first.
    """
    if tree:
        yield from _onehopgreedy(M, seed)
    else:
        for p, i in _onehopgreedy(M, seed):
            yield p

def _onehopgreedy(M, seed = None, alpha = 1/3):
    """
    Return an iterator that yields `(point, index)` pairs, where `point`
    is the next point in a one-hop greedy permutation and `index` is the index
    of the nearest predecessor.

    The optional `seed` parameter indicates the point that should appear first.
    """
    # If no seed is provided, use the first point.
    T = GreedyTree(M)
    G = ClusterGraph(M, seed or next(iter(M)))
    H = G.heap
    root = H.findmax()
    # Yield the first point.
    yield root.center, None

    # Store the indices of the previous points.
    index = {root : 0}

    for i in range(1, len(M)):
        cluster = H.findmax()
        x = max(cluster, key = cluster.dist)
        r = x.dist(cluster.center)
        ra = r * alpha
        potentials = (
            (q,n)
            for n in G.nbrs(cluster)
            for q in n
            if x.dist(q) < ra
        )
        def rangesample(qn):
            q, n = qn
            # TODO: replace with rangecount ?
            return len(T.range(q, ra, ra/10))
        point, parent = max(potentials, key = rangesample)
        newcluster = G.addcluster(point, parent)
        index[newcluster] = i
        yield point, index[parent]



from collections import defaultdict
from greedypermutation.clustergraph import ClusterGraph

def knnsample(M, k, seed = None):
    # If no seed is provided, use the first point.
    G = ClusterGraph(M, seed, 2, 1)
    markednbrs = defaultdict(set)
    H = G.heap
    root = H.findmax()

    # Yield the first point.
    yield root.center

    for i in range(1, len(M)):
        cluster = H.findmax()
        point = cluster.pop()
        radius = 2 * point.dist(cluster.center)

        nearbypts = {q for nbr in G.nbrs(cluster)
                       for q in nbr
                       if q.dist(point) <= radius
                    }

        nearbymarkedpts = {q for q in markednbrs.get(point, ())
                             if q.dist(point) <= radius
                          }

        # If there are fewer than k nearby points, we mark it.
        # If there are at least k, we yield it and add the cluster.
        if len(nearbypts) + len(nearbymarkedpts) < k:
            for p in nearbypts:
                # This is overkill.  Should only add this if point is less than
                # twice the distance to RNN(p)
                markednbrs[p].add(point)
            # In greedy, `addcluster` updates the heap after moving points.
            # Here, we have to do it manually.
            H.changepriority(cluster)
        else:
            if point in markednbrs:
                del markednbrs[point]
            newcluster = G.addcluster(point, cluster)
            yield point



from collections import defaultdict
from greedypermutation.clarksongreedy import greedy
from greedypermutation.maxheap import MaxHeap
from greedypermutation.knnheap import KNNHeap

class Bunch:
    """
    A `Bunch` is a point with a `radius` and a `weight`.
    It stands in for a collection of points.

    The main way to access the points in a bunch is via the `pop` method.
    This returns a bunch and updates the current bunch.

    The main use is to be placed in a heap ordered by radius.
    By storing iterators, the child lists of the tree are not duplicated.
    """
    def __init__(self, node):
        self.node = node
        self.index = 0
        # self.point = node.point
        # self.children = iter(node.children)
        self.weight = node.weight
        # self.radii = iter(node.radii)
        # self.radius = next(self.radii)

    @property
    def point(self):
        return self.node.point

    def radius(self):
        return self.node.radii[self.index]

    def isempty(self):
        return self.weight == 1

    def pop(self):
        if self.index < len(self.node.children):
            child = self.node.children[self.index]
            self.weight -= child.weight
            self.index += 1
            return Bunch(child)
        else:
            return None

    def __len__(self):
        return self.weight

    def __iter__(self):
        """
        Iterate over all points in the bunch.
        """
        yield self.node.point
        for i in range(self.index, len(self.node.children)):
            yield from self.node.children[i]

class Node:
    """
    GreedyTree `Node`s store the following data.
    - a `point`: the location of the node.
    - a `weight`: the number of points in the subtree rooted at the node.
    - `children`: An ordered list of `Node`s.
    - `radii`: An ordered list of radii.  The ith entry is the radius of the
      subtree rooted at the Node with the first `i-1` children removed.
    """

    def __init__(self, point):
        self.point = point
        self.weight = 1
        self.children = []
        self.radii = [0]

    def update(self):
        """
        Update the weight and the radius list.
        """

        for c in reversed(self.children):
            c.update()
            self.weight = 1 + sum(c.weight for c in self.children)
            self.radii.append(max(self.radii[-1],
                                   c.farthest_descendant(self.point)
                                  ))
        self.radii.reverse()

    @property
    def radius(self):
        return self.radii[0]

    def farthest_descendant(self, point):
        """
        Return the largest distance from `point` to any point in the subtree
        rooted at `self`.

        Attempt to improve efficiency by pruning children that are too close.
        """
        farthest = self.point.dist(point)
        for c in self.children:
            if c.point.dist(point) + c.radius > farthest:
                farthest = max(farthest, c.farthest_descendant(point))
        return farthest

    def __iter__(self):
        """
        Iterate over all points in the bunch.

        Warning.  This iteration is destructive.
        """
        yield self.point
        for child in self.children:
            yield from child

    def __len__(self):
        return self.weight


class GreedyTree:
    """
    The `GreedyTree` encodes a tree of points.
    If the points are ordered according to the greedy permutation, the child of
    a point is its nearest predecessor in the ordering.
    """

    def __init__(self, M, seed = None, scaling = float('inf')):
        """
        Initialize a new greedy tree on the input metric space `M`.
        """
        P = []
        self.ch = defaultdict(list)
        self.scaling = scaling
        alpha = 1 - (1/scaling)
        for p, i in greedy(M, seed, tree = True, alpha = alpha):
            newnode = Node(p)
            P.append(newnode)
            if i is not None:
                P[i].children.append(newnode)
        self.root = P[0]
        self.root.update()

    def heap(self):
        """
        Return a max heap over bunches ordered by radius initialized with
        `root`.
        """
        H = MaxHeap(key = Bunch.radius)
        H.insert(Bunch(self.root))
        return H

    def ann(self, q, eps = 1):
        """
        Return a eps-approximate nearest neighbor to q.
        """
        H = self.heap()
        nbr, q_to_nbr = self.root, q.dist(self.root.point)
        close_enough = (eps-1) / eps

        for child in H:
            maxradius = child.radius()

            # If the nearest neighbor could be a child of child.
            # if not child.isempty() and \
            if child.point.dist(q) < q_to_nbr + maxradius:
                p = child.pop()
                q_to_p = q.dist(p.point)
                # If p is the new nearest neighbors, update (nbr, q_to_nbr).
                if q_to_p < q_to_nbr:
                    nbr, q_to_nbr = p, q_to_p
                # If p is a sufficiently close ANN, return it.
                if maxradius < close_enough * q_to_p:
                    return nbr.point
                H.insert(p)
                H.insert(child) # Put the child back in the heap.
        return nbr.point

    def nn(self, q):
        """
        Return the nearest neighbor of q in the GreedyTree.
        """
        H = self.heap()
        nbr, q_to_nbr = self.root, q.dist(self.root.point)

        for child in H:
            # If the nearest neighbor could be a child of child.
            if child.point.dist(q) - child.radius() < q_to_nbr:
                p = child.pop()
                q_to_p = q.dist(p.point)
                # If p is the new nearest neighbors, update (nbr, q_to_nbr).
                if q_to_p < q_to_nbr:
                    nbr, q_to_nbr = p, q_to_p
                H.insert(p)
                H.insert(child) # Put the child back in the heap.
        return nbr.point

    def _range(self, center, radius, slack = 0):
        """
        Produce a minimal collection of bunches that covers the range and is
        contained in the slack range.
        """
        H = self.heap()
        for child in H:
            distance_to_center = child.point.dist(center)
            maxdistance = distance_to_center + child.radius()
            mindistance = distance_to_center - child.radius()
            if maxdistance <= radius + slack:
                yield child
            elif mindistance <= radius:
                p = child.pop()
                if p is not None:
                    H.insert(child)
                    H.insert(p)

    def rangecount(self, center, radius, slack = 0):
        """
        Count the points in the ball with the given `center` and `radius`.

        If the `slack` is nonzero then the count could include some points
        within distance `radius + slack` of the `center`.
        """
        bunches = self._range(center, radius, slack)
        return sum(bunch.weight for bunch in bunches)

    def range(self, center, radius, slack = 0):
        """
        Iterate over the points in `ball(center, radius)`.

        If `slack` is nonzero, some points from `ball(center, radius + slack)`.
        """
        bunches = self._range(center, radius, slack)
        for bunch in bunches:
            yield from bunch

    def range_simple(self, center, radius):
        H = self.heap()
        viable = lambda b: center.dist(b.point) <= b.radius() + radius

        for bunch in H:
            newbunch = bunch.pop()
            if newbunch is None:
                yield bunch.point
            else:
                if viable(bunch):
                    H.insert(bunch)
                if viable(newbunch):
                    H.insert(newbunch)


    def knn(self, k, q):
        H = self.heap()
        # nbr, q_to_nbr = self.root, q.dist(self.root.point)
        knndist = float('inf')
        KNN = KNNHeap(k)
        KNN.insert(self.root, q.dist(self.root.point) + self.root.radius)

        for bunch in H:
            # Inside Split
            if bunch in KNN:
                newbunch = bunch.pop()
                bunch_radius = min(oldradius, bunch.radius())
                KNN.changepriority(bunch, )
            # else:

        return list(KNN)


    def farthest(self, q):
        H = self.heap()
        viable = lambda b, d: q.dist(b.point) + b.radius() > d

        farthest, dist_to_farthest = self.root.point, q.dist(self.root.point)
        for bunch in H:
            newbunch = bunch.pop()
            dist_to_newbunch = q.dist(newbunch.point)
            if dist_to_newbunch > dist_to_farthest:
                dist_to_farthest = dist_to_newbunch
                farthest = newbunch.point
            if viable(newbunch, dist_to_farthest):
                H.insert(newbunch)
            if viable(bunch, dist_to_farthest):
                H.insert(bunch)
        return farthest, dist_to_farthest

    def __iter__(self, greedy = True):
        if greedy:
            H = self.heap()
            yield self.root.point, None

            for child in H:
                if not child.isempty():
                    q = child.pop()
                    yield q.point, child.point
                    H.insert(child)
                    H.insert(q)
        else:
            yield from self.root

    def __str__(self):
        return str(self.ch)

    def __len__(self):
        return len(self.root)



import pytest
from greedypermutation import Point
from greedypermutation.onehopgreedy import onehopgreedy
from metricspaces import MetricSpace

def test_points_on_a_line():
    P = [Point([i]) for i in range(100)]
    GP = list(onehopgreedy(MetricSpace(P), P[50]))
    assert(GP[0] == P[50])
    expected = [50, 16, 83]
    n = len(expected)
    assert(GP[:n] == [P[i] for i in expected])



if __name__ == '__main__':
    pytest.main()



class Point:
    """A simple class to describe points in Euclidean space with float
    coordinates.
    """

    def __init__(self, coords):
        self._p = tuple(float(x) for x in coords)

    @staticmethod
    def fromstring(s):
        """Return a new point object from a string representation.
        """
        return Point(x for x in s.split())

    def dist(self, other):
        """Compute the Euclidean distance between the `self` and `other`.
        If the dimensions don't match the distance is computed in projection
        down to the common subspace.
        """
        return (sum((a-b)**2 for (a,b) in zip(self, other))) ** (0.5)

    def __getitem__(self, index):
        return self._p[index]

    def __eq__(self, other):
        return self._p == other._p

    def __hash__(self):
        return hash(self._p)

    def __str__(self):
        return " ".join(str(c) for c in self._p)

    def __iter__(self):
        return iter(self._p)

    def __len__(self):
        return len(self._p)

    def __repr__(self):
        return str(self)



import unittest
import os
from click.testing import CliRunner
from  greedypermutation.cli import cli

class TestCLI(unittest.TestCase):
    def testgreedy(self):
        runner = CliRunner()
        with runner.isolated_filesystem():
            infile = 'pointfile.txt'
            outfile = 'testoutput.txt'
            with open(infile, 'w') as f:
                f.write(POINTS)
            result = runner.invoke(cli, [infile, '--outfile', outfile])
            self.assertEqual(result.exit_code, 0)
            self.assertTrue(os.path.exists(outfile))
            with open(outfile, 'r') as f:
                for line in f.readlines():
                    self.assertEqual(len(line.split(';')), 2)

    def testgreedy_tostdout(self):
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('pointfile', 'w') as f:
                f.write(POINTS)
            result = runner.invoke(cli, ['pointfile'])
            self.assertEqual(result.exit_code, 0)
            gp = result.output.split('\n')
            self.assertEqual(len(gp[0].split(';')), 2)
            self.assertTrue(';' in gp[0])
            self.assertEqual(gp[0].split(';')[1], 'None')

    def testgreedy_notree(self):
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('pointfile', 'w') as f:
                f.write(POINTS)
            result = runner.invoke(cli, ['pointfile', '--notree'])
            self.assertEqual(result.exit_code, 0)
            gp = result.output.split('\n')
            self.assertTrue(';' not in gp[0])

    def testgreedy_quadratic(self):
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('pointfile', 'w') as f:
                f.write(POINTS)
            result = runner.invoke(cli, ['pointfile', '--notree', '--algorithm', 'quadratic'])
            self.assertEqual(result.exit_code, 0)
            gp = result.output.split('\n')
            self.assertTrue(';' not in gp[0])



POINTS = """\
1 2
100 2
45 3
25 0
"""


if __name__ == '__main__':
    unittest.main()



from ds2.graph import Graph
from greedypermutation.maxheap import MaxHeap


class Cluster:
    def __init__(self, center):
        """
        Create a new cluster with the given `center`.

        A new cluster only contains a single point, its center.
        """
        self.points = {center}
        self.center = center
        self.radius = 0

    def addpoint(self, p):
        """
        Add the point `p` to the cluster.
        """
        self.points.add(p)
        self.radius = max(self.radius, self.dist(p))

    def dist(self, point):
        """
        Return the distance between the center of the cluster and `point`.
        Note, this allows the cluster to be treated almost like a point.
        """
        return self.center.dist(point)

    def updateradius(self):
        """
        Set the radius of the cluster to be the farthest distance from a point
        to the center.
        """
        self.radius = max((self.dist(p) for p in self.points), default = 0)

    def pop(self):
        """
        Remove and return the farthest point in the cluster.

        Returns `None` if there there are no points other than the center.
        """
        if len(self) == 1:
            return None
        p = max(self.points, key = self.dist)
        # This is linear time!  We should maybe use a heap here.
        # However, the pop is followed by an update that will iterate over all
        # the points anyway.
        # For knn sampling, the pop might not require such an iteration.
        self.points.remove(p)
        self.updateradius()
        return p

    def __len__(self):
        """
        Return the total number of points in the cluster, including the center.
        """
        return len(self.points)

    def __iter__(self):
        """
        Return an iterator over the points in the cluster.
        """
        return iter(self.points)

    def __contains__(self, point):
        """
        Return True if and only if `point` is in the cluster.
        """
        return point in self.points

    def __lt__(self, other):
        """
        Clusters are ordered by their radii.
        """
        return self.radius > other.radius

    def __repr__(self):
        return str(self.center)

class ClusterGraph(Graph):
    Vertex = Cluster

    # Initialize it as an empty graph.
    def __init__(self, points, root = None, nbrconstant = 1, moveconstant = 1):
        """
        Initialize a new ClusterGraph.

        It starts with an iterable of points.  The first point will be the
        center of the default cluster and all other points will be placed
        inside.

        There are two constants that can be set.
        The first `nbrconstant`, which controls the distance between neighbors.
        The second is `moveconstant` which determines when a point is moved
        when a new cluster is formed.  The default value for both constants is
        `1`.  This moves a point whenever it has a new nearest neighbor

        The theoretical guarantees are only valid when
        `moveconstant <= nbrconstant`.  As a result, setting these any other
        way raises an exception.
        """
        # Initialize the `ClusterGraph` to be a `Graph`.
        super().__init__()
        P = iter(points)

        if nbrconstant < moveconstant:
            raise RuntimeError("The move constant must not be larger than the"
                               "neighbor constant.")
        self.nbrconstant = nbrconstant
        self.moveconstant = moveconstant
        # Make a cluster to start the graph.  Use the first point as the root
        # if none is give.
        root_cluster = self.Vertex(root or next(P))
        # Add the points to the root cluster.
        # It doesn't matter if the root point is also in the list of points.
        # It will not be added twice.
        for p in P:
            root_cluster.addpoint(p)
        # Add the new cluster as the one vertex of the graph.
        self.addvertex(root_cluster)
        self.addedge(root_cluster, root_cluster)
        self.heap = MaxHeap([root_cluster], key = lambda c: c.radius)

    def iscloseenoughto(self, p, q):
        return q.dist(p.center) <= p.radius + q.radius + \
                      self.nbrconstant * max(p.radius, q.radius)

    def addcluster(self, newcenter, parent):
        """
        Add a new cluster centered at `newcenter`.

        The `parent` is a suffciently close cluster that is already in the
        graph.
        It is used to find nearby clusters to be the neighbors.
        The clusters are rebalanced with points moving from nearby clusters into
        the new cluster if it is closer.
        """
        # Create the new cluster.
        newcluster = self.Vertex(newcenter)
        # Make the cluster a new vertex.
        self.addvertex(newcluster)
        self.addedge(newcluster, newcluster)

        # Rebalance the new cluster.
        for nbr in self.nbrs(parent):
            self.rebalance(newcluster, nbr)
            self.heap.changepriority(nbr)

        # Add neighbors to the new cluster.
        for newnbr in self.nbrs_of_nbrs(parent):
            if self.iscloseenoughto(newcluster, newnbr):
                self.addedge(newcluster, newnbr)

        # After all the radii are updated, prune edges that are too long.
        for nbr in set(self.nbrs(parent)):
            self.prunenbrs(nbr)

        self.heap.insert(newcluster)
        return newcluster

    def pop(self):
        cluster = self.heap.findmax()
        point = cluster.pop
        self.heap.changepriority(cluster)
        return point

    def rebalance(self, a, b):
        """
        Move points from the cluster `b` to the cluster `a` if they are
        sufficiently closer to `a.center`.
        """
        points_to_move = {p for p in b.points
                            if a.dist(p) < self.moveconstant * b.dist(p)}
        b.points -= points_to_move
        for p in points_to_move:
            a.addpoint(p)
        # The radius of self (`a`) is automatically updated by addpoint.
        # The other radius needs to be manually updated.
        b.updateradius()

    def nbrs_of_nbrs(self, u):
        return {b for a in self.nbrs(u) for b in self.nbrs(a)}

    def prunenbrs(self, u):
        """
        Eliminate neighbors that are too far with respect to the current
        radius.
        """
        nbrs_to_delete = set()
        for v in self.nbrs(u):
            if not self.iscloseenoughto(u,v):
                nbrs_to_delete.add(v)

        # Prune the excess edges.
        for v in nbrs_to_delete:
            self.removeedge(u,v)



from greedypermutation import GreedyTree
from greedypermutation.greedytree import Node, Bunch
from metricspaces import MetricSpace, NumpyPoint as Point
import unittest


class TestGreedyTree(unittest.TestCase):
    def setUp(self):
        self.P = [Point([c]) for c in [0, 100, 49, 25, 60, 12, 81]]
        self.M = MetricSpace(self.P)

    def testinit(self):
        P = self.P
        T = GreedyTree(self.M, P[0])
        # It should have 7 points
        self.assertEqual(len(T), len(self.P))
        # It should have a root that is a node.
        self.assertTrue(isinstance(T.root, Node))
        # The point in the root node should be P[0].
        self.assertEqual(T.root.point, P[0])
        # The root has 3 children (100, 49, and 12)
        self.assertEqual(len(T.root.children), 3)
        # The first child is 100.
        onehundred = T.root.children[0]
        self.assertEqual(onehundred.point, P[1])
        # The second child is 49.
        fortynine = T.root.children[1]
        self.assertEqual(fortynine.point, P[2])
        # The third child is 12.
        twelve = T.root.children[2]
        self.assertEqual(twelve.point, P[5])
        # Twelve has no children
        self.assertEqual(len(twelve.children), 0)
        # There are only two nodes in the subtree at onehundred.
        self.assertEqual(len(onehundred), 2)
        # The lone child of 100 is 81.
        self.assertEqual(onehundred.children[0].point, P[6])

    def testNN(self):
        P = self.P
        T = GreedyTree(self.M)
        self.assertEqual(T.nn(Point([11])), P[5])
        self.assertEqual(T.nn(Point([5])), P[0])
        self.assertEqual(T.nn(Point([51])), P[2])
        self.assertEqual(T.nn(Point([75])), P[6])

    def testANN_eps_equals_zero(self):
        P = self.P
        T = GreedyTree(self.M)
        self.assertEqual(T.ann(Point([11])), P[5])
        self.assertEqual(T.ann(Point([5])), P[0])
        self.assertEqual(T.ann(Point([51])), P[2])
        self.assertEqual(T.ann(Point([75])), P[6])

    def testnodeiter(self):
        T = GreedyTree(self.M, next(iter(self.M)))
        self.assertEqual(set(T.root), set(self.M))


if __name__ == '__main__':
    unittest.main()



from ds2viz.element import Group, Line
from greedypermutation.maxheap import MaxHeap
from greedypermutation.vizpoint import VizPoint

class VizGreedyTree(Group):
    def __init__(self, T):
        super().__init__()
        P = set()
        for p, q in T:
            P.add(p)
            if q is not None:
                self.addelement(Line(p, q))
        for p in P:
            self.addelement(VizPoint(*p))



from greedypermutation.knnheap import KNNHeap
from greedypermutation.greedytree import Bunch
import unittest


class TestKNNHeap(unittest.TestCase):
    def setup(self):
        print("running a test")

    def testinit(self):
        H = KNNHeap(3)

    def testinsert_simple(self):
        H = KNNHeap(5)

        # B = Bunch()
        # H.insert()

if __name__ == '__main__':
    unittest.main()



import click
import os
from greedypermutation import Point
from metricspaces import MetricSpace

@click.command()
@click.argument('pointsfile', type = click.File('r'))
# @click.option('--outfile', type = click.Path(allow_dash = True),
@click.option('--outfile', type = click.File('w'),
                default = '-',
                help = "Where to store the output.")
@click.option('--algorithm',
                type=click.Choice(['clarkson', 'quadratic'], case_sensitive=False),
                default = 'clarkson',
                help='Which algorithm to use: `quadratic` or `clarkson`.')
@click.option('--tree/--notree', default = True,
                help='Include the entire tree in the output or not.')
def cli(pointsfile, outfile, algorithm, tree):
    """
    Compute a greedy permutation of the points in the `pointsfile`.

    The output will be saved to the file specified by `outfile`.

    By default it will run the Clarkson algorithm.
    If `--algorithm quadratic` is specified, then the quadratic algorithm will
    be used.

    The `--tree` and `--notree flags determine if the tree is included in the
    output.  This is managed by appending the index of the predecessor after
    a semicolon.
    """
    M = MetricSpace()
    M.fromstrings(pointsfile.readlines(), Point.fromstring)

    if algorithm == 'quadratic':
        import greedypermutation.quadraticgreedy as algo
    else:
        import greedypermutation.clarksongreedy as algo
    if tree:
        for p, i in algo.greedy(M, tree = True):
            outfile.write(str(p) + ';' + str(i) + '\n')
    else:
        for p in algo.greedy(M):
            outfile.write(str(p) + '\n')



import unittest
from greedypermutation import Point

class TestPoint(unittest.TestCase):
    def test_init_2D_points(self):
        # Just checking that init doesn't crash
        Point([0,0])
        Point([-1,2])
        Point([-1.1, 2.99])

    def testlen(self):
        P = [Point([1]), Point([1,2]), Point([1,2,3])]
        self.assertEqual(len(P[0]), 1)
        self.assertEqual(len(P[1]), 2)
        self.assertEqual(len(P[2]), 3)

    def test_init_with_other_iterables(self):
        p = Point(range(100))
        self.assertEqual(len(p), 100)

    def test_dist_1D(self):
        P = [Point([i]) for i in range(10)]
        self.assertEqual(P[1].dist(P[5]), 4)
        self.assertEqual(P[5].dist(P[1]), 4)
        self.assertEqual(P[0].dist(P[9]), 9)
        self.assertEqual(P[1].dist(P[1]), 0)
        self.assertEqual(P[6].dist(P[8]), 2)

    def test_dist_2D(self):
        a = Point([1,3])
        b = Point([13,8])
        self.assertEqual(a.dist(b), 13)

    def test_dist_with_different_dimensions(self):
        """If two points are in different dimensions, the higher dimensional
        point should be projected to the lower dimensional space.
        """
        p, q = Point([3,4]), Point([0])
        self.assertEqual(p.dist(q), 3)
        r = Point([0,0,100])
        self.assertEqual(p.dist(r), 5)

    def test_eq(self):
        a = Point([3,4])
        b = Point([3.0, 4.0])
        self.assertTrue(a == b)

    def test_hash(self):
        a = Point([30000000000,4])
        b = Point([30000000000.0, 4.0])
        self.assertEqual(hash(a), hash(b))

    def test_fromstring(self):
        p = Point.fromstring("  4  2.01 999")
        self.assertEqual(len(p), 3)
        a = Point.fromstring(' 3  4  12')
        self.assertEqual(len(a), 3)
        b = Point([0,0,0])
        self.assertEqual(a.dist(b), 13)

    def test_iter(self):
        p = Point(3* i for i in range(4))
        for i, c in enumerate(p):
            self.assertEqual(c, 3 * i)

if __name__ == '__main__':
    unittest.main()



import unittest
from random import randrange, seed
from collections import defaultdict
from greedypermutation import (Point,
                               quadraticgreedy,
                               clarksongreedy,
                               )
from metricspaces import MetricSpace

class GreedyTests:
    def testgreedy(self):
        greedy = self.implementation.greedy
        P = [Point([i]) for i in range(3)]
        M = MetricSpace(P)
        gp = list(greedy(M, P[0]))
        self.assertEqual(gp, [P[0], P[2], P[1]])

    def testgreedy_exponential_example(self):
        greedy = self.implementation.greedy
        P = [Point([(-3)**i]) for i in range (100)]
        # print([str(p) for p in P])
        M = MetricSpace(P)
        gp = greedy(M, P[0])
        self.assertEqual(next(gp), P[0])
        self.assertEqual(next(gp), P[99])
        for i in range(98, 0, -1):
            self.assertEqual(next(gp), P[i])

    def testgreedytree_randomexample(self):
        greedy = self.implementation.greedy
        root = Point([0])
        P = MetricSpace([root] + [Point([x]) for x in [8,12, 100, 40, 70, 1, 72]])
        gp = greedy(P, root, tree = True)
        self.assertEqual(next(gp), (Point([0]), None))
        self.assertEqual(next(gp), (Point([100]), 0)) # radius = 100
        self.assertEqual(next(gp), (Point([40]), 0))
        self.assertEqual(next(gp), (Point([70]), 1))
        self.assertEqual(next(gp), (Point([12]), 0))
        self.assertEqual(next(gp), (Point([8]), 4))
        self.assertEqual(next(gp), (Point([72]), 3))

    def testgreedytree_bigexample(self):
        greedy = self.implementation.greedy
        n = 600
        coords = set()
        while len(coords) < n:
            coords.add((randrange(100), randrange(100), randrange(100)))
        P = [Point(c) for c in coords]
        M = MetricSpace(P)

        GP = list(greedy(M, P[0], tree = True))
        radii = [p.dist(GP[i][0]) for p,i in GP if i is not None]
        # Check that the insertion radii are nonincreasing.
        for i in range(n-2):
            self.assertTrue(radii[i] >= radii[i+1], str([i, radii[i], radii[i+1]]))

    def testgreedytree_example2(self):
        greedy = self.implementation.greedy
        P = [Point([c]) for c in [0, 100, 49, 25, 60, 12, 81]]
        M = MetricSpace(P)
        root = P[0]
        gt = list(greedy(M, root, tree = True))
        gp = [p for p, i in gt]
        ch = defaultdict(list)
        for p, i in greedy(M, root, tree = True):
            if i is not None:
                ch[gp[i]].append(p)
        self.assertEqual(gp, [P[i] for i in [0, 1, 2, 3, 6, 5, 4]])

    def testgreedytree_example3(self):
        greedy = self.implementation.greedy
        P = [Point([c]) for c in [0, 1, 3, 5, 20, 30]]
        M = MetricSpace(P)
        gt = list(greedy(M, P[0], tree = True))
        gp = [p for p, i in gt]
        ch = defaultdict(set)
        for p, i in gt:
            if i is not None:
                ch[gp[i]].add(p)

        self.assertEqual(gp, [P[0], P[5], P[4], P[3], P[2], P[1]])
        self.assertEqual(ch[P[0]], {P[5], P[3], P[1]})
        self.assertEqual(ch[P[1]], set())
        self.assertEqual(ch[P[2]], set())
        self.assertEqual(ch[P[3]], {P[2]})
        self.assertEqual(ch[P[4]], set())
        self.assertEqual(ch[P[5]], {P[4]})

def _test(impl):
    class GreedyTestCase(unittest.TestCase, GreedyTests):
        implementation = impl
    return GreedyTestCase


TestQuadraticGreedy = _test(quadraticgreedy)
TestClarksonGreedy = _test(clarksongreedy)

if __name__ == '__main__':
    unittest.main()



from .graph import Graph

class WeightedGraph(Graph):
    def addedge(self, u, v, wt = 1):
        self._nbrs[u][v] = wt
        self._nbrs[v][u] = wt

    def addvertex(self, v):
        self._nbrs.setdefault(v, {})

    def wt(self, u, v):
        return self._nbrs[u][v]

    def dijkstra(self, v):
        tree = {}
        D = {v: 0}
        tovisit = PriorityQueue()
        tovisit.insert((None,v), 0)
        while tovisit:
            a,b = tovisit.removemin()
            if b not in tree:
                tree[b] = a
                if a is not None:
                    D[b] = D[a] + self.wt(a,b)
                for n in self.nbrs(b):
                    tovisit.add((b,n), D[b] + self.wt(b,n))
        return tree, D



from greedypermutation.clustergraph import Cluster, ClusterGraph

def greedy(M, seed = None, tree = False, alpha = 1):
    """
    Return an iterator that yields the points of `M` ordered by a greedy
    permutation.

    The optional `seed` parameter indicates the point that should appear first.

    The optional `alpha` parameter sets the approximation in the
    `ClusterGraph`.  Every point will have a parent that is a `1/alpha`
    approximate nearest neighbor.  The resulting greedy permutation will be a
    `1/alpha` approximation.
    """
    if tree:
        yield from _greedy(M, seed, alpha)
    else:
        for p, i in _greedy(M, seed, alpha):
            yield p

def _greedy(M, seed = None, alpha = 1):
    """
    Return an iterator that yields `(point, index)` pairs, where `point`
    is the next point in a greedy permutation and `index` is the index of they
    nearest predecessor.

    The optional `seed` parameter indicates the point that should appear first.
    """
    # If no seed is provided, use the first point.
    G = ClusterGraph(M, seed or next(iter(M)), alpha, alpha)
    H = G.heap
    root = H.findmax()

    # Yield the first point.
    yield root.center, None

    # Store the indices of the previous points.
    index = {root : 0}

    for i in range(1, len(M)):
        cluster = H.findmax()
        point = cluster.pop()
        newcluster = G.addcluster(point, cluster)
        index[newcluster] = i
        yield point, index[cluster]



from greedypermutation.maxheap import MaxHeap

class KNNHeap:
    def __init__(self, k):
        self.heap = MaxHeap(key = lambda x: x[1])
        self.bunches = {}
        self.k = k
        self._len = 0

    def __contains__(self, bunch):
        return bunch in self.bunches

    def insert(self, bunch, distance):
        """
        Insert a new bunch as long as it doesn't cause the heap to get too
        large.  If the new bunch is closer than other bunches in the heap,
        then some may be removed to make space.

        The invariant maintained by insert is that if the bunches are sorted
        by increasing upper bound distance to the query, then the heap contains
        the minimum prefix that has at least k points.

        The items inserted should be pairs.  This is important as we always
        need the distance (i.e. the priority when we remove an item.)
        """

        if len(self) < self.k:
            self.bunches[bunch] = distance
            self.heap.insert((bunch, distance))
        else:
            top, current_radius = self.heap.findmax()
            if distance < current_radius:
                if len(self) - len(top) + len(bunch) >= self.k:
                    self.heap.removemax()
                    del self.bunches[top]
                self.insert(bunch, distance)

    def splitbunch(self, q, bunch):
        """
        Splits a bunch into two bunches using pop and tries to insert both
        back into the heap.
        """
        b = bunch.pop()
        b_radius = min(q.dist(b.point) + b.radius, self.bunches[bunch])

        a = bunch
        a_radius = min(q.dist(a.point) + a.radius, self.bunches[bunch])

        del self.bunches[a]
        self.insert(a, a_radius)
        self.insert(b, b_radius)

    def __len__(self):
        return self._len

    def __iter__(self):
        return iter(self.heap)



import unittest
from greedypermutation.maxheap import MaxHeap

class TestMaxHeap(unittest.TestCase):
    def testinit(self):
        H = MaxHeap()
        H = MaxHeap({1,2,3})
        H = MaxHeap([1,2,3])
        self.assertEqual(len(H), 3)

    def testfindmax(self):
        H = MaxHeap([1,2,5,4,3])
        self.assertEqual(H.findmax(), 5)

    def testremovemax(self):
        H = MaxHeap([1,2,5,4,3])
        self.assertEqual(H.removemax(), 5)
        self.assertEqual(len(H), 4)
        self.assertEqual(H.removemax(), 4)
        self.assertEqual(len(H), 3)
        self.assertEqual(H.removemax(), 3)
        H.insert(9)
        H.insert(0)
        self.assertEqual(H.removemax(), 9)
        self.assertEqual(len(H), 3)
        self.assertEqual(H.removemax(), 2)

if __name__ == '__main__':
    unittest.main()
